import pytest

from nbstore.store import Store


@pytest.fixture(scope="module", autouse=True)
def _execute(store: Store):
    return store.execute("seaborn.ipynb")


def test_outputs(store: Store):
    outputs = store.get_outputs("seaborn.ipynb", "fig:seaborn")
    assert isinstance(outputs, list)
    assert len(outputs) == 1
    assert isinstance(outputs[0], dict)
    assert outputs[0]["output_type"] == "execute_result"
    assert "text/plain" in outputs[0]["data"]


def test_data(store: Store):
    data = store.get_data("seaborn.ipynb", "fig:seaborn")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["text/plain"].startswith("%% Creator: Matplotlib,")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import nbformat

from .image import create_image_file, get_content

if TYPE_CHECKING:
    from nbformat import NotebookNode


class Store:
    notebook_dir: Path
    notebooks: dict[Path, NotebookNode]
    st_mtime: dict[Path, float]
    current_path: Path | None

    def __init__(self, notebook_dir: Path | str) -> None:
        self.notebook_dir = Path(notebook_dir)
        self.notebooks = {}
        self.st_mtime = {}
        self.current_path = None

    def _read(self, abs_path: Path) -> NotebookNode:
        mtime = abs_path.stat().st_mtime

        if (nb_ := self.notebooks.get(abs_path)) and self.st_mtime[abs_path] == mtime:
            return nb_

        nb: NotebookNode = nbformat.read(abs_path, as_version=4)  # type: ignore

        self.notebooks[abs_path] = nb
        self.st_mtime[abs_path] = mtime

        return nb

    def _write(self, abs_path: Path) -> None:
        if nb := self.notebooks.get(abs_path):
            nbformat.write(nb, abs_path)

    def _get_abs_path(self, url: str) -> Path:
        if not url and self.current_path:
            return self.current_path

        abs_path = (self.notebook_dir / url).absolute()

        if abs_path.exists():
            self.current_path = abs_path
            return abs_path

        msg = "Unknown path."
        raise ValueError(msg)

    def get_notebook(self, url: str) -> NotebookNode:
        abs_path = self._get_abs_path(url)
        return self._read(abs_path)

    def get_cell(self, url: str, identifier: str) -> dict[str, Any]:
        nb = self.get_notebook(url)
        return get_cell(nb, identifier)

    def get_source(self, url: str, identifier: str) -> str:
        nb = self.get_notebook(url)
        return get_source(nb, identifier)

    def get_outputs(self, url: str, identifier: str) -> list:
        nb = self.get_notebook(url)
        return get_outputs(nb, identifier)

    def get_stream(self, url: str, identifier: str) -> str | None:
        outputs = self.get_outputs(url, identifier)
        return get_stream(outputs)

    def get_data(self, url: str, identifier: str) -> dict[str, str]:
        outputs = self.get_outputs(url, identifier)
        return get_data(outputs)

    def add_data(self, url: str, identifier: str, mime: str, data: str) -> None:
        outputs = self.get_outputs(url, identifier)
        if output := get_data_by_type(outputs, "display_data"):
            output[mime] = data

    def save_notebook(self, url: str) -> None:
        self._write(self._get_abs_path(url))

    def delete_data(self, url: str, identifier: str, mime: str) -> None:
        outputs = self.get_outputs(url, identifier)
        output = get_data_by_type(outputs, "display_data")
        if output and mime in output:
            del output[mime]

    def get_language(self, url: str) -> str:
        nb = self.get_notebook(url)
        return get_language(nb)

    def execute(self, url: str) -> NotebookNode:
        from nbconvert.preprocessors import ExecutePreprocessor

        nb = self.get_notebook(url)
        ep = ExecutePreprocessor(timeout=600)
        ep.preprocess(nb)
        return nb

    def get_content(self, url: str, identifier: str) -> tuple[str | bytes, str] | None:
        data = self.get_data(url, identifier)
        return get_content(data)

    def create_image_file(
        self,
        url: str,
        identifier: str,
        filename: Path | str,
        *,
        delete: bool = False,
    ) -> Path | None:
        data = self.get_data(url, identifier)
        return create_image_file(data, filename, delete=delete)


def get_cell(nb: NotebookNode, identifier: str) -> dict[str, Any]:
    for cell in nb["cells"]:
        source: str = cell["source"]
        if source.startswith(f"# #{identifier}\n"):
            return cell

    msg = f"Unknown identifier: {identifier}"
    raise ValueError(msg)


def get_source(nb: NotebookNode, identifier: str) -> str:
    if source := get_cell(nb, identifier).get("source", ""):
        source = "\n".join(source.split("\n")[1:])

    return source


def get_outputs(nb: NotebookNode, identifier: str) -> list:
    return get_cell(nb, identifier).get("outputs", [])


def get_data_by_type(outputs: list, output_type: str) -> dict[str, str] | None:
    for output in outputs:
        if output["output_type"] == output_type:
            if output_type == "stream":
                return {"text/plain": output["text"]}

            return output["data"]

    return None


def get_stream(outputs: list) -> str | None:
    if data := get_data_by_type(outputs, "stream"):
        return data["text/plain"]

    return None


def get_data(outputs: list) -> dict[str, str]:
    for type_ in ["display_data", "execute_result"]:
        if data := get_data_by_type(outputs, type_):
            return data

    msg = "No output data"
    raise ValueError(msg)


def get_language(nb: dict) -> str:
    return nb["metadata"]["kernelspec"]["language"]



import pytest

from nbstore.store import Store


def test_get_data_none(store: Store):
    with pytest.raises(ValueError, match="No output data"):
        store.get_data("pgf.ipynb", "fig:none")


def test_add_data(store: Store):
    from nbstore.store import get_data_by_type

    url = "add.ipynb"
    identifier = "fig:add"
    mime = "mime"
    data = "data"

    assert mime not in store.get_data(url, identifier)

    store.add_data(url, identifier, mime, data)

    assert mime in store.get_data(url, identifier)
    store.save_notebook(url)

    assert mime in store.get_data(url, identifier)

    outputs = store.get_outputs(url, identifier)
    output = get_data_by_type(outputs, "display_data")
    assert output
    store.delete_data(url, identifier, mime)
    store.save_notebook(url)

    assert mime not in store.get_data(url, identifier)


def test_get_language(store: Store):
    assert store.get_language("add.ipynb") == "python"


def test_get_cell_error(store: Store):
    with pytest.raises(ValueError, match="Unknown identifier: fig:invalid"):
        store.get_data("pgf.ipynb", "fig:invalid")


def test_current_path(store: Store):
    a = store.get_source("pgf.ipynb", "fig:stream")
    b = store.get_source("", "fig:stream")
    assert a == b


def test_abs_path_error(store: Store):
    with pytest.raises(ValueError, match="Unknown path."):
        store.get_source("unknown.ipynb", "fig:stream")


def test_stream_none():
    from nbstore.store import get_stream

    assert get_stream([]) is None



import pytest

from nbstore.store import Store


@pytest.fixture(scope="module", autouse=True)
def _execute(store: Store):
    return store.execute("pgf.ipynb")


def test_cell(store: Store):
    cell = store.get_cell("pgf.ipynb", "fig:pgf")
    assert isinstance(cell, dict)
    assert "cell_type" in cell


def test_source(store: Store):
    source = store.get_source("pgf.ipynb", "fig:pgf")
    assert isinstance(source, str)
    assert "plot" in source


def test_outputs(store: Store):
    outputs = store.get_outputs("pgf.ipynb", "fig:pgf")
    assert isinstance(outputs, list)
    assert len(outputs) == 1
    assert isinstance(outputs[0], dict)
    assert outputs[0]["output_type"] == "display_data"
    assert "text/plain" in outputs[0]["data"]


def test_data(store: Store):
    data = store.get_data("pgf.ipynb", "fig:pgf")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["text/plain"].startswith("%% Creator: Matplotlib,")


def test_stream(store: Store):
    assert store.get_stream("pgf.ipynb", "fig:stream") == "123\n"



import pytest
from PIL import Image

from nbstore.store import Store


@pytest.fixture(scope="module", autouse=True)
def _execute(store: Store):
    return store.execute("png.ipynb")


def test_cell(store: Store):
    cell = store.get_cell("png.ipynb", "fig:png")
    assert isinstance(cell, dict)
    assert "cell_type" in cell


def test_source(store: Store):
    source = store.get_source("png.ipynb", "fig:png")
    assert isinstance(source, str)
    assert "plot" in source


def test_outputs(store: Store):
    outputs = store.get_outputs("png.ipynb", "fig:png")
    assert isinstance(outputs, list)
    assert len(outputs) == 2
    assert isinstance(outputs[0], dict)
    assert outputs[0]["output_type"] == "execute_result"
    assert "text/plain" in outputs[0]["data"]
    assert isinstance(outputs[1], dict)
    assert outputs[1]["output_type"] == "display_data"


def test_data(store: Store):
    data = store.get_data("png.ipynb", "fig:png")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["image/png"].startswith("iVBO")


def test_content(store: Store):
    data = store.get_content("png.ipynb", "fig:png")
    assert isinstance(data, tuple)
    assert len(data) == 2
    assert isinstance(data[0], bytes)
    assert data[1] == ".png"


def test_image(store: Store):
    file = store.create_image_file("png.ipynb", "fig:png", "a", delete=True)
    assert file
    assert file.exists()
    assert file.name == "a.png"
    image = Image.open(file)
    assert image.format == "PNG"
    assert image.size == (1136, 826)


def test_image_none():
    from nbstore.image import create_image_file

    assert create_image_file({}, "a") is None



from __future__ import annotations

import atexit
import base64
from pathlib import Path


def get_content(data: dict[str, str]) -> tuple[str | bytes, str] | None:
    if text := data.get("application/pdf"):
        return base64.b64decode(text), ".pdf"

    for mime, text in data.items():
        if mime.startswith("image/"):
            ext = mime.split("/")[1]
            return base64.b64decode(text), f".{ext}"

    return None


def create_image_file(
    data: dict[str, str],
    filename: Path | str,
    *,
    delete: bool = False,
) -> Path | None:
    decoded = get_content(data)

    if decoded is None:
        return None

    content, suffix = decoded
    file = Path(filename).with_suffix(suffix)

    if isinstance(content, str):
        file.write_text(content)
    else:
        file.write_bytes(content)

    if delete:
        atexit.register(lambda: file.unlink(missing_ok=True))

    return file



from __future__ import annotations

import atexit
import base64
import shutil
import tempfile
from pathlib import Path

from .formatter import RASTER_BEGIN, RASTER_END


def split(text: str) -> tuple[str, dict[str, str]]:
    if not text.endswith(RASTER_END):
        return text, {}

    text, image = text.split(f"\n{RASTER_BEGIN}\n", maxsplit=1)
    images = image.splitlines()[:-1]

    image_dict = {}
    for image in images:
        name, imagetext = image.split(":", maxsplit=1)
        image_dict[name] = imagetext

    return text, image_dict


def convert(text: str) -> str:
    text, images = split(text)
    if not images:
        return text

    name, _ = next(iter(images.keys())).rsplit("-img", maxsplit=1)
    tempdir = create_temp_dir()
    text = text.replace(name, f"{tempdir.as_posix()}/{name}")

    for name, imagetext in images.items():
        data = base64.b64decode(imagetext)
        path = tempdir / name
        path.write_bytes(data)

    return text


def create_temp_dir() -> Path:
    dirname = tempfile.mkdtemp()
    path = Path(dirname)
    atexit.register(lambda: shutil.rmtree(path))
    return path



import io
import re
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from IPython.lib.pretty import RepresentationPrinter
from PIL import Image

from nbstore.formatter import RASTER_BEGIN, RASTER_END


@pytest.fixture(scope="module")
def text():
    from nbstore.formatter import matplotlib_figure_to_pgf

    fig, axes = plt.subplots(1, 2, figsize=(4, 2))
    for i in [0, 1]:
        data = np.random.randn(10, 10)
        axes[i].imshow(data, interpolation="nearest", aspect=1)
        axes[i].set(xlabel="x", ylabel="α")

    out = io.StringIO()
    rp = RepresentationPrinter(out)

    matplotlib_figure_to_pgf(fig, rp, None)
    return out.getvalue()


def test_matplotlib_figure_to_pgf_raster(text: str):
    assert f"\n{RASTER_BEGIN}\n" in text
    assert text.endswith(f"\n{RASTER_END}")


def test_split(text: str):
    from nbstore.pgf import split

    text, images = split(text)
    assert text.startswith("%% Creator: Matplotlib, PGF backend")
    assert text.endswith("\\endgroup%\n")

    assert len(images) == 2

    for name, data in images.items():
        assert name in text
        assert data.startswith("iVBOR")
        assert not data.endswith("\n")


def test_split_none():
    from nbstore.pgf import split

    text, images = split("abc")
    assert text == "abc"
    assert not images


def test_convert_pgf_text(text: str):
    from nbstore.pgf import convert

    text = convert(text)

    for k, x in enumerate(re.findall(r"\\includegraphics\[.+?\]{(.+?)}", text)):
        assert x.endswith(f"-img{k}.png")
        assert Path(x).exists()
        image = Image.open(x)
        assert image.format == "PNG"
        assert image.size == (141, 141)


def test_convert_none():
    from nbstore.pgf import convert

    assert convert("abc") == "abc"
